"""
Copyright (c) 2008, Carl Scheffler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the copyright owner may not be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

def optimize_paths(paths):
    def nearest_neighbour(points, index):
        """Find the nearest neighbour to points[index] in points.
        It is assumed that points is sorted by x"""
        import math

        radius = 1e10
        result = -1
        i = 0
        signs = {-1: index > 0, 1: index < len(points)-1}
        while signs[-1] or signs[1]:
            i += 1
            for sgn in signs:
                if signs[sgn]:
                    if sgn*points[index+sgn*i]['x'] - sgn*points[index]['x'] < radius:
                        check = math.hypot(points[index]['x']-points[index+sgn*i]['x'],
                                           points[index]['y']-points[index+sgn*i]['y'])
                        if check < radius:
                            radius = check
                            result = index+sgn*i

                        if index+sgn*i in [0, len(points)-1]:
                            signs[sgn] = False
                    else:
                        signs[sgn] = False

        return result

    points = []
    for i in range(len(paths)):
        points.append({'x': paths[i]['path'][0][0],
                       'y': paths[i]['path'][0][1],
                       'curve': i,
                       'dir': 1})
        points.append({'x': paths[i]['path'][-1][0],
                       'y': paths[i]['path'][-1][1],
                       'curve': i,
                       'dir': -1})
    points.sort(cmp = lambda a, b: cmp(a['x'], b['x']))

    new_paths = []
    index = 0
    while len(points) > 1:
        new_paths.append(paths[points[index]['curve']])
        if points[index]['dir'] < 0: # reverse
            new_paths[-1]['path'].reverse()

        for i in range(len(points)):
            if (points[i]['curve'] == points[index]['curve']) and \
               (points[i]['dir'] != points[index]['dir']):
                new_index = i
                break
        del points[index]
        if index < new_index: new_index -= 1
        index = new_index

        if len(points) > 1:
            new_index = nearest_neighbour(points, index)
            del points[index]
            if index < new_index: new_index -= 1
            index = new_index
    return new_paths

def optimize_epi(commands):
    from convert import epi2paths, paths2epi

    start = commands.find('%1BIN;')
    if start == -1:
        start = 0
    else:
        start += 6
    stop  = commands.rfind('%0B')
    if stop == -1:
        stop = len(commands)
    else:
        stop -= 1
    header = commands[:start]
    footer = commands[stop:]
    commands = commands[start:stop]

    return header + paths2epi(optimize_paths(epi2paths(commands))) + footer

def optimize_camm(commands):
    from convert import camm2paths, paths2camm

    start = commands.find('!ST1;')
    if start == -1:
        start = None
    else:
        start += 5

    header = commands[:start]
    commands = commands[start:]

    return header + paths2camm(optimize_paths(camm2paths(commands)))
