"""
Copyright (c) 2008, Carl Scheffler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the copyright owner may not be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import division
from scipy import *

def render_rml(commands, tool=0.125, profile='flat'):
    import re

    def bound(bounding_box, point):
        if point[0] < bounding_box[0]:
            bounding_box[0] = point[0]
        if point[0] > bounding_box[1]:
            bounding_box[1] = point[0]
        if point[1] < bounding_box[2]:
            bounding_box[2] = point[1]
        if point[1] > bounding_box[3]:
            bounding_box[3] = point[1]
        if len(point) > 2:
            if point[2] < bounding_box[4]:
                bounding_box[4] = point[2]
            if point[2] > bounding_box[5]:
                bounding_box[5] = point[2]


    DPI = 40*25.4 # 40 dots / mm

    position = [0., 0.]
    bounding_box = [1e10, -1e10, 1e10, -1e10, 1e10, -1e10]

    start = commands.find('!MC1;') + 5
    stop  = commands.rfind('!MC0')

    commands = commands[start:stop-1].strip()
    commands = re.split(';\s*', commands)

    # Find bounding box
    for command in commands:
        if command[0] == 'P':
            if command[1] == 'D':
                bound(bounding_box, position)
            temp = command[2:].split(',')
            position[0] = float(temp[0])/DPI
            position[1] = float(temp[1])/DPI
            if command[1] == 'D':
                bound(bounding_box, position)
        elif command[0:3] == '!PZ':
            temp = command[3:].split(',')
            z_down = float(temp[0])/DPI
            z_up   = float(temp[1])/DPI
            if z_down < bounding_box[4]:
                bounding_box[4] = z_down
            if z_up > bounding_box[5]:
                bounding_box[5] = z_up
        elif command[0] == 'Z':
            temp = command[1:].split(',')
            assert len(temp)%3 == 0, "Number of coordinates (x,y,z) not a multiple of 3."
            for i in range(0, len(temp), 3):
                position = [float(temp[i])/DPI, float(temp[i+1])/DPI, float(temp[i+2])/DPI]
                bound(bounding_box, position)
    print bounding_box

    # Generate grid
    def transform(pos, bounding_box, pixel_size):
        return ((pos[0]-(bounding_box[0]-tool/2))/pixel_size,
                (pos[1]-(bounding_box[2]-tool/2))/pixel_size,
                pos[2])

    def update_grid(pos0, pos1, grid_z, stencil, bounding_box, pixel_size):
        #print pos0, pos1, bounding_box
        stencil_radius = (stencil.shape[0]-1)//2

        # change to pixel coordinate system
        pos0 = transform(pos0, bounding_box, pixel_size)
        pos1 = transform(pos1, bounding_box, pixel_size)

        # round to nearest pixel
        pixel0 = [int(round(pos0[0])), int(round(pos0[1]))]
        pixel1 = [int(round(pos1[0])), int(round(pos1[1]))]

        # check if line is vertical
        if (pixel0[0] == pixel1[0]) and (pixel0[1] == pixel1[1]):
            start0 = min(max(pixel0[0]-stencil_radius, 0), grid_z.shape[0])
            stop0 = max(min(pixel0[0]+stencil_radius+1, grid_z.shape[0]), 0)
            start1 = min(max(pixel0[1]-stencil_radius, 0), grid_z.shape[1])
            stop1 = max(min(pixel0[1]+stencil_radius+1, grid_z.shape[1]), 0)
            
            grid_z[start0:stop0, start1:stop1] = minimum(grid_z[start0:stop0, start1:stop1], stencil[start0-pixel0[0]+stencil_radius:stop0-pixel0[0]+stencil_radius, start1-pixel0[1]+stencil_radius:stop1-pixel0[1]+stencil_radius]+min(pos0[2],pos1[2]))
            return

        major = (abs(pixel1[0]-pixel0[0]) < abs(pixel1[1]-pixel0[1]))
        minor = 1-major

        delta_major = (pixel0[major] < pixel1[major])*2-1
        delta_minor = (pos1[minor]-pos0[minor])/(pos1[major]-pos0[major])
        delta_z = (pos1[2]-pos0[2])/(pos1[major]-pos0[major])

        pixel = [None, None]
        for pixel[major] in xrange(pixel0[major], pixel1[major]+delta_major, delta_major):
            pixel[minor] = int(round(pos0[minor] + delta_minor * (pixel[major]-pos0[major])))
            z = pos0[2] + delta_z * (pixel[major]-pos0[major])

            start0 = min(max(pixel[0]-stencil_radius, 0), grid_z.shape[0])
            stop0 = max(min(pixel[0]+stencil_radius+1, grid_z.shape[0]), 0)
            start1 = min(max(pixel[1]-stencil_radius, 0), grid_z.shape[1])
            stop1 = max(min(pixel[1]+stencil_radius+1, grid_z.shape[1]), 0)

            #print grid_z[start0:stop0, start1:stop1].shape
            #print stencil[start0-pixel[0]+stencil_radius:stop0-pixel[0]+stencil_radius, start1-pixel[1]+stencil_radius:stop1-pixel[1]+stencil_radius].shape
            #print start0, stop0, start1, stop1
            #print pixel, stencil_radius
            #print pixel0, pixel1
            #print pos0, pos1

            grid_z[start0:stop0, start1:stop1] = minimum(grid_z[start0:stop0, start1:stop1], stencil[start0-pixel[0]+stencil_radius:stop0-pixel[0]+stencil_radius, start1-pixel[1]+stencil_radius:stop1-pixel[1]+stencil_radius]+z)
        

    pixels = 512
    pixel_size = max(bounding_box[1]-bounding_box[0]+tool, bounding_box[3]-bounding_box[2]+tool)/pixels
    x_pixels = int(ceil((bounding_box[1]-bounding_box[0]+tool)/pixel_size))
    y_pixels = int(ceil((bounding_box[3]-bounding_box[2]+tool)/pixel_size))

    grid_z = ones((x_pixels, y_pixels), dtype=float) * bounding_box[5]

    stencil_size = int(floor(tool/pixel_size/2)*2+1)
    x = linspace(-pixel_size*(stencil_size-1)/2, pixel_size*(stencil_size-1)/2, stencil_size)
    sqr_dist = outer(x, ones(stencil_size))**2 + outer(ones(stencil_size), x)**2
    not_in_stencil = sqr_dist > (tool/2)**2
    if profile == 'flat':
        stencil = not_in_stencil * 1e10
    elif profile == 'round':
        stencil = (tool/2-sqrt((tool/2)**2-sqr_dist*(1-not_in_stencil))) + 1e10*not_in_stencil
    else:
        raise ValueError, "Unknown profile '%s'"%profile

    position = [0., 0., 0.]
    pen_height = {'up': 0., 'down': 0.}
    for command in commands:
        if command[0] == 'P':
            height = pen_height[{'U': 'up', 'D': 'down'}[command[1]]]
            old_position = position[:2]
            temp = command[2:].split(',')
            position = [float(temp[0])/DPI, float(temp[1])/DPI, height]
            update_grid(old_position + [height], position, grid_z, stencil, bounding_box, pixel_size)
        elif command[0:3] == '!PZ':
            temp = command[3:].split(',')
            pen_height['down'] = float(temp[0])/DPI
            pen_height['up']   = float(temp[1])/DPI
        elif command[0] == 'Z':
            temp = command[1:].split(',')
            assert len(temp)%3 == 0, "Number of coordinates (x,y,z) not a multiple of 3."
            for i in range(0, len(temp), 3):
                old_position = position[:]
                position = [float(temp[i])/DPI, float(temp[i+1])/DPI, float(temp[i+2])/DPI]
                update_grid(old_position, position, grid_z, stencil, bounding_box, pixel_size)

    minmax = [min(ravel(grid_z)), max(ravel(grid_z))]

    import Image
    im = Image.fromarray(array((grid_z-minmax[0])/(minmax[1]-minmax[0])*255, dtype=uint8).transpose(), mode='L')
    fp = file('render_rml.png', 'wb')
    im.save(fp)
    fp.close()

    return None

import sys

fp = file(sys.argv[1], 'rt')
rml = fp.read()
fp.close()

render_rml(rml, tool=1/64, profile='flat')
